#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>
#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <cglm/cglm.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"
#include "proj.h"
#include "entities.h"
#include "generation.h"
#include "display.h"

int flashlight = 10 ;

int MAX_SIZE = 8192 ;

int* drawOrder ;

void init_draworder() {
    drawOrder = malloc(sizeof(int)*6) ;
}

float vertices[9] ;
void gl_renderTriangle(unsigned int shaderProgram, unsigned int VAO, unsigned int VBO, double x0, double y0, double z0, double x1, double y1, double z1, double x2, double y2, double z2, float r, float g, float b) {
    vertices[0] = (float)x0;
    vertices[1] = (float)y0;
    vertices[2] = (float)z0;
    vertices[3] = (float)x1;
    vertices[4] = (float)y1;
    vertices[5] = (float)z1;
    vertices[6] = (float)x2;
    vertices[7] = (float)y2;
    vertices[8] = (float)z2;

    glBindBuffer(GL_ARRAY_BUFFER, VBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // note that this is allowed, the call to glVertexAttribPointer registered VBO as the vertex attribute's bound vertex buffer object so afterwards we can safely unbind
    glBindBuffer(GL_ARRAY_BUFFER, 0); 

    // You can unbind the VAO afterwards so other VAO calls won't accidentally modify this VAO, but this rarely happens. Modifying other
    // VAOs requires a call to glBindVertexArray anyways so we generally don't unbind VAOs (nor VBOs) when it's not directly necessary.
    glBindVertexArray(0); 

    glUseProgram(shaderProgram);
    glBindVertexArray(VAO);
    glDrawArrays(GL_TRIANGLES, 0, 3);
}

double near = 0.4 ;
double far = 30.0 ;

double top = 1.0 ;
double bottom = -1.0 ;
double left = -1.0 ;
double right = 1.0 ;

double z_depth(double z) {
    return ((2.0*(1/z - 1/near)/(1/far - 1/near)) - 1.0);
    //return ((1/z - 1/near)/(1/far - 1/near));
    //return (2*z -1);
}

pt_2d gl_project(pt_2d p) {
    double w = -p.z;
    pt_2d res = {
        .x = p.x*(2.0*near/(right-left)) + p.z*(right+left)/(right-left),
        .y = p.y*(2.0*near/(top-bottom)) + p.z*(top+bottom)/(top-bottom),
        .z = -p.z*(far+near)/(far-near) -(2*far*near)/(far-near),
    };
    res.x = res.x / w;
    res.y = res.y / w;
    res.z = res.z / w;
    return res;
}

mat4 model, view, projection;
void gl_renderSurface(unsigned int shaderProgram, unsigned int fragmentShader, unsigned int VAO, unsigned int VBO, cube_0* c, int sf) {
    float vertices[] = {
        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),
        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),
        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),
        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),

        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),
        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),

        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),
        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),
        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),

        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),
        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),
        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),

        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),
        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(-c->d/2.0),
        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(-c->d/2.0),

        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),
        (float)(c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),
        (float)(-c->w/2.0), (float)(-c->h/2.0), (float)(c->d/2.0),
        (float)(-c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),
        (float)(c->w/2.0), (float)(c->h/2.0), (float)(c->d/2.0),
    };

    glm_mat4_identity(model);
    glm_translate(model, (vec3){(float)(c->x+c->w/2.0), (float)(c->y+c->h/2.0), (float)(c->z+c->d/2.0)});
    glm_rotate(model, (float)(-c->hz_angle), (vec3){0.0f, 1.0f, 0.0f});
    glm_rotate(model, (float)(c->vt_angle), (vec3){1.0f, 0.0f, 0.0f});

    vec3 dir0;
    vec3 direction;

    dir0[0] = sinf((float)(rot_hz)) * cosf((float)(rot_vt));
    dir0[1] = -sinf((float)(rot_vt));
    dir0[2] = cosf((float)(rot_hz)) * cosf((float)(rot_vt));

    glm_vec3_normalize(dir0); // Normalize to unit length
    glm_vec3_add((vec3){(float)camx, (float)camy, (float)camz}, dir0, direction);

    glm_mat4_identity(view);
    glm_lookat((vec3){(float)camx, (float)camy, (float)camz}, direction, (vec3){0.0f, 1.0f, 0.0f}, view);

    glm_perspective(glm_rad((float)fov), 1500.0f / 1000.0f, 0.1f, 100.0f, projection);

    glUniformMatrix4fv(glGetUniformLocation(shaderProgram, "model"), 1, GL_FALSE, (float*)model);
    glUniformMatrix4fv(glGetUniformLocation(shaderProgram, "view"), 1, GL_FALSE, (float*)view);
    glUniformMatrix4fv(glGetUniformLocation(shaderProgram, "projection"), 1, GL_FALSE, (float*)projection);

    glUseProgram(fragmentShader);
    glUniform4f(glGetUniformLocation(fragmentShader, "u_color"), c->red/255.0f, c->green/255.0f, c->blue/255.0f, 1.0f);

    glBindBuffer(GL_ARRAY_BUFFER, VBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // note that this is allowed, the call to glVertexAttribPointer registered VBO as the vertex attribute's bound vertex buffer object so afterwards we can safely unbind
    glBindBuffer(GL_ARRAY_BUFFER, 0); 

    // You can unbind the VAO afterwards so other VAO calls won't accidentally modify this VAO, but this rarely happens. Modifying other
    // VAOs requires a call to glBindVertexArray anyways so we generally don't unbind VAOs (nor VBOs) when it's not directly necessary.
    glBindVertexArray(0); 

    glUseProgram(shaderProgram);
    glBindVertexArray(VAO);
    glDrawArrays(GL_TRIANGLES, 0, 36);
}

void gl_renderCube(unsigned int shaderProgram, unsigned int fragmentShader, unsigned int VAO, unsigned int VBO, cube_0* c) {
    for(int k = 0; k < 6; k++) {
        gl_renderSurface(shaderProgram, fragmentShader, VAO, VBO, c, k);
    }
}

void gl_renderAll(unsigned int shaderProgram, unsigned int fragmentShader, unsigned int VAO, unsigned int VBO) {
    for(int k = 0; k < current_room->map_size; k++) {
        gl_renderCube(shaderProgram, fragmentShader, VAO, VBO, current_room->map[k]);
    }
    for(int k = 0; k < current_room->tps_size; k++) {
        gl_renderCube(shaderProgram, fragmentShader, VAO, VBO, current_room->tps[k]->hitbox);
    }
    for(int k = 0; k < current_room->ent_len; k++) {
        gl_renderCube(shaderProgram, fragmentShader, VAO, VBO, current_room->ents[k].pos);
    }
}